{-----------------------------------------------------------------------------}
{ Labels handling interface part.                                (c) 1992 YGP }

  const _id_global = $01;                     { local constant must be global }
        _id_used =   $02;                                     { label is used }

  type id_tree = array [0..4] of pointer;         {     }

       id_types = (id_undeclared,             { undeclared forward used label }
		   id_forward,                 { forward declaration of label }
                   id_expression,                     { unresolved expression }
                   id_address,                        { byte relative address }
                   id_pg_addr,                        { page relative address }
		   id_a_addr,                              { alligned address }
                   id_eword,           { external identifier (range 0..65535) }
		   (*
		   id_ebyte,             { external identifier (range 0..255) }
		   id_byte,                                    { range 0..255 }
                   id_nbyte,                                 { range -1..-256 }
		   *)
                   id_word,                                  { range 0..65535 }
                   id_nword,                               { range -1..-65536 }
                   id_lint,                              { full longint range }
                   id_real,                                { real (FP) number }
                   id_defined);      { identifier for conditional compilation }

       idp = ^identifier;
       ex_lstp = ^ex_lst;

       val_kinds = (v_err,
                    v_none,

                    v_byte,                { 0..255 (    ) }
                    v_word,              { 0..65535 (    ) }
                    v_negw,            { -65536..-1 (    ) }
                    v_cnst,            { any constant (except string or real) }
                    v_real,                                   { real constant }

                    v_str,                                  { string constant }
                    v_addr,                      { address (relocatable word) }
                    v_extr,                                   { external word }

                    v_op_add,                                             { + }
                    v_op_sub,                                             { - }
                    v_op_mul,                                             { * }
                    v_op_div,                                             { / }
                    v_op_mod,                                             { \ }
                    v_op_and,                                             { & }
                    v_op_or,                                              { | }
		    v_op_xor,                                             { ^ }
		    v_op_lt,            				  { < }
		    v_op_gt,  						  { > }
		    v_op_le,  						 { <= }
		    v_op_ge, 						 { >= }
		    v_op_ne,					 	 { <> }
		    v_op_eq,						  { = }

                    v_op_not,                                             { ! }
                    v_op_neg,                                        { negate }

                    v_hi_bt,                                        { hi byte }
                    v_lo_bt,                                        { lo byte }
                    v_hi_wd,                                        { hi word }
                    v_lo_wd,                                        { lo word }
                    v_w_swap,                                     { swap word }
                    v_l_swap,                                 { swap longword }

		    v_expr,                                      { expression }
		    v_fwrd);                         { forward declared label }

       valuep = ^value;
       expr_ptr = ^expr;
       expr = packed record
         expr: valuep;
         line: cardinal;                                { line Nr in source }
         f_no: shortcard;                                         { file no }
         kind: byte;                                 { see _e_xxx constants }
         case boolean of
           false: (offs: cardinal);                        { offset in code }
           true:  (eidp: idp);
       end { expr_list };

       value = packed record case v_kind: val_kinds of
         v_fwrd: (id: idp);                          { forward declared label }
         v_cnst: (l:  longint);     { longint constant includes all subranges }
         v_real: (r:  real);                                   { fp constants }

         v_str:  (s:  stringp);                               { ptr to string }
         v_byte: (b:  byte);
         v_negw,
         v_word: (w:  word);
         v_expr: (case boolean of
                    false: (e: valuep);
                    true:  (ex: expr_ptr));
         v_op_and,
	 v_op_sub: (l_op, r_op: valuep);               { left / right operand }

	 v_op_neg: (s_op: valuep);                           { single operand }
       end { value };

       ex_lst = record
         expr: expr_ptr;
         next: ex_lstp;
       end { ex_lst };

       identifier = packed record
         id_flags: byte;                             { see _id_xxx constants }
	 case id_type: id_types of
	   id_undeclared,
	   id_expression,
	   id_forward: (elst: ex_lstp; rsrv: pointer {internal use only} );
	   (*
	   id_byte,
	   id_nbyte,
	   id_ebyte:   (bval: shortcard);
	   *)
	   id_pg_addr,
	   id_address,
	   id_word,
	   id_nword,
	   id_eword:   (wval: cardinal);
	   id_a_addr:  (_wval: cardinal; allgn: cardinal);
	   id_lint:    (lval: longint);
	   id_real:    (rval: real);
	   id_defined: ();
       end { identifier };

       level_ptr = ^level_rec;
       level_rec = record
         prev: level_ptr;                  { ,     }
         labs: id_tree;               {       }
         pnam: ident_str;                              {    }
       end { level_rec };

  var cc_defined:  id_tree;      {      }
      cc_def_cfg:  id_tree;     {   -,    init }
      global_labs: id_tree;                              {   }
      expr_labs:   id_tree;           {       }
      flvl,             {      (nil => global_labs) }
      lvl:         level_ptr;  {    (nil   proc) }
                                                                            (*)
      value_sizes: packed array [val_kinds] of shortcard;                   (*)

  function find_lab(nmsz: shortcard;
                    (*)const{*)var{*} nm: packed_name): idp;   (*) forward; (*)
    {    (label)    level,     ..}
    {    (    ).                }

  function search_lab(nmsz: shortcard;
                      (*)const{*)var{*} nm: packed_name;
                      var t: id_tree): idp;                    (*) forward; (*)
    {    (label)   . }

  function insert_lab(nmsz: shortcard;
                      (*)const{*)var{*} nm: packed_name;
                      (*)const{*)var{*} id: identifier;
                      var t: id_tree): idp;                    (*) forward; (*)
    {    (label)   . }

  procedure dispose_tree(var t: id_tree; prev: level_ptr;
                         cc_def: boolean);     (*) forward; (*)
    {    . }

  procedure check_tree(unresolved: boolean);                   (*) forward; (*)
    {     (forward, expression) . }

  procedure change_lab(fidp: idp;
                       (*)const{*)var{*} id: identifier);      (*) forward; (*)
    {   forward  expression     }

  procedure get_idname(fidp: idp; var name: ident_str);        (*) forward; (*)
    {-     }

  procedure PrintXref(t: id_tree);                             (*) forward; (*)

  {              (  }
  {  Xref.         .    }
  {           .      }
  {    xxx.xrf           }
  {     .      . }
  {     ,       Xref    . }

{-----------------------------------------------------------------------------}
